==> v1.3.1.zip: database-attribute_system/Database-Attribute_System/ClassAction.cs <==
﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Text;

namespace eu.railduction.netcore.dll.Database_Attribute_System
{
    public class ClassAction
    {
        /// <summary>
        /// Fills an given dbObject with given data<para/>
        /// Data-attribute-names and class-fieldNames have to match! (non case-sensitive)
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="classObject">Given object (marked with Db-attributes)</param>
        /// <param name="data">The data</param>
        /// <param name="runDataLossChecks">This checks if any class-field and data-attribute does not exists in either (Slower)</param>
        public static void FillObject<T>(T classObject, Dictionary<string, object> data, bool runDataLossChecks = true)
        {
            Type classType = classObject.GetType();

            string tableName = Function.GetDbTableName(classType);

            // Get class-fields
            Dictionary<string, FieldInfo> dbFields = Function.ReadDbClassFields(classType);

            if (runDataLossChecks)
            {
                // Check every data-attribute for match in class-fields
                foreach (KeyValuePair<string, object> data_keySet in data)
                {
                    bool isFound = false;
                    foreach (KeyValuePair<string, FieldInfo> field_keySet in dbFields)
                    {
                        if (field_keySet.Key.ToLower() == data_keySet.Key.ToLower())
                            isFound = true;
                    }

                    if (!isFound)
                        throw new InvalidOperationException($"Could not fill object. Data-Attribute '{data_keySet.Key}' was not found in class!");
                }
                // Check every class-field for match in data-attributes
                foreach (KeyValuePair<string, FieldInfo> field_keySet in dbFields)
                {
                    bool isFound = false;
                    foreach (KeyValuePair<string, object> data_keySet in data)
                    {
                        if (field_keySet.Key.ToLower() == data_keySet.Key.ToLower())
                            isFound = true;
                    }

                    if (!isFound)
                        throw new InvalidOperationException($"Could not fill object. Class-field '{field_keySet.Key}' was not found in data!");
                }
            }      

            // Iterate through data
            foreach (KeyValuePair<string, object> data_keySet in data)
            {
                // Interate through class-fields
                foreach (KeyValuePair<string, FieldInfo> field_keySet in dbFields)
                { 
                    // If its a match, set the value
                    if (field_keySet.Key.ToLower() == data_keySet.Key.ToLower())
                    {
                        field_keySet.Value.SetValue(classObject, data_keySet.Value);
                        break;
                    }
                }
            }
        }


        /// <summary>
        /// Resolves an object with the database<para/>
        /// Needs to have primaryKey/s-value/s set!<para/>
        /// - Generates an query<para/>
        /// - Sends an query via Func<para/>
        /// - Fills the object with data
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="classObject">Given object (marked with Db-attributes)</param>
        /// <param name="queryExecutor">Function to handle query-calls - Has to return Dictionary[attributeName, attributeValue]</param>
        /// <param name="runDataLossChecks">This checks if any class-field and data-attribute does not exists in either (Slower)</param>
        public static void ResolveByPrimaryKey<T>(T classObject, Func<string, List<Dictionary<string, object>>> queryExecutor, bool runDataLossChecks = true)
        {
            string query = QueryBuilder.SelectByPrimaryKey(classObject);   // Generate query
            List<Dictionary<string, object>> dataSet = queryExecutor(query);    // Execute
            FillObject(classObject, dataSet[0], runDataLossChecks);   // Fill the object
        }

        /// <summary>
        /// Gets a list of dbObjects by attribute/s
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="classType">Type of class</param>
        /// <param name="fields">class-fields for select</param>
        /// <param name="queryExecutor">Function to handle query-calls - Has to return Dictionary[attributeName, attributeValue]</param>
        /// <param name="runDataLossChecks">This checks if any class-field and data-attribute does not exists in either (Slower)</param>
        /// <returns></returns>
        public static List<T> GetListByAttribute<T>(Type classType, Dictionary<string, object> fields, Func<string, List<Dictionary<string, object>>> queryExecutor, bool runDataLossChecks = true) where T : new()
        {
            string tableName = Function.GetDbTableName(classType);  // Get database-tableName

            Function.ConvertAttributeToDbAttributes(classType, fields);

            string query = QueryBuilder.SelectByAttribute(tableName, fields);   // Generate query
            List<Dictionary<string, object>> dataSet = queryExecutor(query);    // Execute

            List<T> objs = new List<T>() { };
            foreach(Dictionary<string, object> data in dataSet)
            {
                T obj = new T();    // New object
                FillObject(obj, data, runDataLossChecks);   // Fill it
                objs.Add(obj);      // Add to list
            }

            return objs;    // Return list
        }
    }
}

==> v1.4.1.zip: database-attribute_system/Database-Attribute_System/ClassAction.cs <==
﻿using eu.railduction.netcore.dll.Database_Attribute_System.Attributes;
using System;
using System.Collections.Generic;
using System.Reflection;
using System.Text;

namespace eu.railduction.netcore.dll.Database_Attribute_System
{
    public class ClassAction
    {
        /// <summary>
        /// Fills an given dbObject with given data<para/>
        /// Data-attribute-names and class-fieldNames have to match! (non case-sensitive)
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="classObject">Given object (marked with Db-attributes)</param>
        /// <param name="data">The data</param>
        /// <param name="runDataLossChecks">This checks if any class-field and data-attribute does not exists in either (Slower)</param>
        public static void FillObject<T>(T classObject, Dictionary<string, object> data, bool runDataLossChecks = true)
        {
            Type classType = classObject.GetType();

            string tableName = Function.GetDbTableName(classType);

            // Get class-fields
            Dictionary<string, FieldInfo> dbFields = Function.ReadDbClassFields(classType);

            if (runDataLossChecks)
            {
                // Check every data-attribute for match in class-fields
                foreach (KeyValuePair<string, object> data_keySet in data)
                {
                    bool isFound = false;
                    foreach (KeyValuePair<string, FieldInfo> field_keySet in dbFields)
                    {
                        if (field_keySet.Key.ToLower() == data_keySet.Key.ToLower())
                            isFound = true;
                    }

                    if (!isFound)
                        throw new InvalidOperationException($"Could not fill object. Data-Attribute '{data_keySet.Key}' was not found in class!");
                }
                // Check every class-field for match in data-attributes
                foreach (KeyValuePair<string, FieldInfo> field_keySet in dbFields)
                {
                    bool isFound = false;
                    foreach (KeyValuePair<string, object> data_keySet in data)
                    {
                        if (field_keySet.Key.ToLower() == data_keySet.Key.ToLower())
                            isFound = true;
                    }

                    if (!isFound)
                        throw new InvalidOperationException($"Could not fill object. Class-field '{field_keySet.Key}' was not found in data!");
                }
            }      

            // Iterate through data
            foreach (KeyValuePair<string, object> data_keySet in data)
            {
                // Interate through class-fields
                foreach (KeyValuePair<string, FieldInfo> field_keySet in dbFields)
                { 
                    // If its a match, set the value
                    if (field_keySet.Key.ToLower() == data_keySet.Key.ToLower())
                    {
                        field_keySet.Value.SetValue(classObject, data_keySet.Value);
                        break;
                    }
                }
            }
        }


        /// <summary>
        /// Gets an dbObject by primaryKey/s
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="classObject">Given object (marked with Db-attributes)</param>
        /// <param name="queryExecutor">Function to handle query-calls - Has to return Dictionary[attributeName, attributeValue]</param>
        /// <param name="runDataLossChecks">This checks if any class-field and data-attribute does not exists in either (Slower)</param>
        public static T GetByPrimaryKey<T>(Type classType, object primaryKeyValue, Func<string, List<Dictionary<string, object>>> queryExecutor, bool runDataLossChecks = true) where T : new()
        {
            Dictionary<string, object> primaryKeyData = new Dictionary<string, object>() { };
            primaryKeyData.Add(null, primaryKeyValue);

            return GetByPrimaryKey<T>(classType, primaryKeyData, queryExecutor, runDataLossChecks);
        }
        public static T GetByPrimaryKey<T>(Type classType, string primaryKeyName, object primaryKeyValue, Func<string, List<Dictionary<string, object>>> queryExecutor, bool runDataLossChecks = true) where T : new()
        {
            Dictionary<string, object> primaryKeyData = new Dictionary<string, object>() { };
            primaryKeyData.Add(primaryKeyName, primaryKeyValue);

            return GetByPrimaryKey<T>(classType, primaryKeyData, queryExecutor, runDataLossChecks);
        }
        public static T GetByPrimaryKey<T>(Type classType, Dictionary<string, object> primaryKeyData, Func<string, List<Dictionary<string, object>>> queryExecutor, bool runDataLossChecks = true) where T: new()
        {
            // Create new empty object
            T obj = new T();

            // Read all fields
            Dictionary<string, FieldInfo> dbFields = Function.ReadDbClassFields(classType);
            // iterate thru them to check and fill object
            foreach (KeyValuePair<string, FieldInfo> field in dbFields)
            {
                // primaryKeys
                if (field.Value.GetCustomAttribute(typeof(DbPrimaryKey), true) is DbPrimaryKey pkey)
                {
                    bool dataMatchFound = false;

                    // Now search the corresponding primaryKeyData
                    foreach (KeyValuePair<string, object> primaryKey in primaryKeyData)
                    {
                        // primaryKey matches
                        if(field.Value.Name.ToLower() == primaryKey.Key.ToLower())
                        {
                            // Set data
                            field.Value.SetValue(obj, primaryKey.Value);

                            dataMatchFound = true;
                            break;
                        }
                    }

                    // If no data was found matching this field
                    if (!dataMatchFound) throw new InvalidOperationException($"Cannot create object with primaryKeyData. No data assigned to field '{field.Value.Name}'");
                }
            }

            string query = QueryBuilder.SelectByPrimaryKey(obj);   // Generate query
            List<Dictionary<string, object>> dataSet = queryExecutor(query);    // Execute
            FillObject(obj, dataSet[0], runDataLossChecks);   // Fill the object

            return obj;
        }

        /// <summary>
        /// Gets a list of dbObjects by attribute/s
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="classType">Type of class</param>
        /// <param name="fields">class-fields for select</param>
        /// <param name="queryExecutor">Function to handle query-calls - Has to return Dictionary[attributeName, attributeValue]</param>
        /// <param name="runDataLossChecks">This checks if any class-field and data-attribute does not exists in either (Slower)</param>
        /// <returns></returns>
        public static List<T> GetListByAttribute<T>(Type classType, Dictionary<string, object> fields, Func<string, List<Dictionary<string, object>>> queryExecutor, bool runDataLossChecks = true) where T : new()
        {
            string tableName = Function.GetDbTableName(classType);  // Get database-tableName

            Function.ConvertAttributeToDbAttributes(classType, fields);

            string query = QueryBuilder.SelectByAttribute(tableName, fields);   // Generate query
            List<Dictionary<string, object>> dataSet = queryExecutor(query);    // Execute

            List<T> objs = new List<T>() { };
            foreach(Dictionary<string, object> data in dataSet)
            {
                T obj = new T();    // New object
                FillObject(obj, data, runDataLossChecks);   // Fill it
                objs.Add(obj);      // Add to list
            }

            return objs;    // Return list
        }



        /// <summary>
        /// Resolves all foreignKeys with the database<pragma/>
        /// Only works if the foreignKey is single (not assembled)!
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="classObject">Given object (marked with Db-attributes)</param>
        /// <param name="queryExecutor">Function to handle query-calls - Has to return Dictionary[attributeName, attributeValue]</param>
        /// <param name="max_depth">Determents how deep resolving will be executed</param>
        /// <param name="runDataLossChecks">This checks if any class-field and data-attribute does not exists in either (Slower)</param>
        public static void ResolveForeignKeys<T>(T classObject, Func<string, List<Dictionary<string, object>>> queryExecutor, int max_depth = 1, bool runDataLossChecks = true) where T: new()
        {
            Type classType = classObject.GetType();

            // Get class-fields
            Dictionary<string, FieldInfo> dbFields = Function.ReadDbClassFields(classType);

            foreach (KeyValuePair<string, FieldInfo> dbField in dbFields)
            {
                // If field is foreignKey
                if (dbField.Value.GetCustomAttribute(typeof(DbForeignKey), true) is DbForeignKey fkey)
                {
                    FieldInfo f_Field = fkey._foreignKeyField;
                    object f_value = f_Field.GetValue(classObject);

                    // When its empty, get it
                    if(f_value == null)
                    {
                        f_value = GetByPrimaryKey<T>(classType, f_value, queryExecutor, runDataLossChecks); ;
                    }

                    // Recursive resolving
                    if (max_depth - 1 > 0)
                    {
                        ResolveForeignKeys(f_value, queryExecutor, max_depth - 1, runDataLossChecks);
                    }
                }
            }
        }
    }
}
